import torch
from torch import optim
from torch.nn import functional as F
import pickle
import copy


def save_torch_model(model, optimizer_class, loss_fn_class, save_path):
    """
    Save a PyTorch model with its structure but without optimizer or loss function state.
    """
    
    # Deep copy the model structure without weights
    new_model = copy.deepcopy(model)

    # Reset parameters to ensure no weights are copied
    def reset_parameters(layer):
        if hasattr(layer, 'reset_parameters'):
            layer.reset_parameters()

    new_model.apply(reset_parameters)

    # Store only the class of the optimizer and loss function, not the state
    optimizer_info = {
        'class_name': optimizer_class.__name__,  # Store the class of the optimizer
    }

    loss_info = {
        'class_name': loss_fn_class.__name__,  # Store the class of the loss function
    }

    # Create the data dictionary to serialize
    model_data = {
        'model': new_model,
        'optimizer': optimizer_info,
        'loss': loss_info,
    }

    # Serialize and save the model data
    with open(save_path, 'wb') as f:
        pickle.dump(model_data, f)


def load_torch_model(ginger_path):
    """
    Load a PyTorch model from a serialized .ginger file, re-creating the optimizer and loss function.
    """
    
    # Load the model data from the serialized file
    with open(ginger_path, 'rb') as f:
        model_data = pickle.load(f)

    # Extract the model structure
    model = model_data['model']

    # Re-create the optimizer from the class name
    optimizer_class_name = model_data['optimizer']['class_name']
    optimizer_class = getattr(optim, optimizer_class_name, optim.SGD)  # Default to SGD
    optimizer = optimizer_class(model.parameters())  # Create a new optimizer instance

    # Re-create the loss function from the class name
    loss_fn_class_name = model_data['loss']['class_name']
    loss_fn_class = getattr(F, loss_fn_class_name, torch.nn.CrossEntropyLoss)  # Default to CrossEntropyLoss
    loss_fn = loss_fn_class()  # Create a new loss function instance

    return model, optimizer, loss_fn
